from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
)
import itertools as it

from cached_property import cached_property

from .abc import (
    Model,
)

if TYPE_CHECKING:
    from typing import (
        Generator,
        Tuple,
        Any,
        Optional,
        Iterable,
        List,
        Set,
    )
    from .positions import (
        Position,
    )
    from .planned_trips import (
        PlannedTrip
    )
    from .vehicles import (
        Vehicle,
    )

logger = logging.getLogger(__name__)


class Stop(Model):
    __slots__ = [
        'vehicle',
        'position',
        'pickups',
        'deliveries',
        'previous',
    ]
    vehicle: Vehicle
    position: Position
    previous: Optional[Stop]
    pickups: Set[PlannedTrip, ...]
    deliveries: Set[PlannedTrip, ...]

    def __init__(self, vehicle: Vehicle, position: Position, previous: Optional[Stop],
                 pickups: Set[PlannedTrip, ...] = None, deliveries: Set[PlannedTrip, ...] = None):

        if pickups is None:
            pickups = set()
        if deliveries is None:
            deliveries = set()

        self.vehicle = vehicle
        self.position = position

        self.pickups = pickups
        self.deliveries = deliveries

        self.previous = previous

    @property
    def identifier(self) -> str:
        trips_sequence = ''.join(
            it.chain(
                (f'P{planned_trip.trip_identifier}' for planned_trip in self.pickups),
                (f'D{planned_trip.trip_identifier}' for planned_trip in self.deliveries),
            )
        )
        return ','.join((
            f'{self.position}',
            f'{self.arrival_time:.2f}:{self.departure_time:.2f}',
            f'({trips_sequence})',
        ))

    @property
    def planned_trips(self) -> Iterable[PlannedTrip]:
        yield from self.pickups
        yield from self.deliveries

    @property
    def trips(self):
        yield from (planned_trip.trip for planned_trip in self.planned_trips)

    @property
    def capacity(self) -> float:
        result = sum(trip.capacity for trip in self.planned_trips)
        assert 0 <= result
        return result

    @property
    def all_previous(self) -> List[Stop]:
        if self.previous is None:
            return []
        return [self.previous] + self.previous.all_previous

    def append_pickup(self, planned_trip: PlannedTrip) -> None:
        assert planned_trip.origin == self.position
        self.pickups.add(planned_trip)

    def append_delivery(self, planned_trip: PlannedTrip) -> None:
        assert planned_trip.destination == self.position
        self.deliveries.add(planned_trip)

    def extend_pickups(self, iterable: Iterable[PlannedTrip]) -> None:
        self.pickups.update(iterable)

    def extend_deliveries(self, iterable: Iterable[PlannedTrip]) -> None:
        self.deliveries.update(iterable)

    @property
    def down_time(self) -> float:
        if not any(self.pickups):
            return 0.0
        return max(pt.down_time for pt in self.pickups)

    @property
    def earliest(self):
        if not any(self.pickups):
            return 0.0
        return max(pt.trip.origin_earliest for pt in self.pickups)

    @property
    def load_time(self) -> float:
        if not any(self.planned_trips):
            return 0.0
        return max(pt.trip.origin_duration for pt in self.planned_trips)

    @property
    def previous_departure_time(self) -> float:
        if self.previous is None:
            return self.vehicle.origin_earliest
        return self.previous.departure_time

    @property
    def previous_position(self) -> Position:
        if self.previous is None:
            return self.vehicle.origin_position
        return self.previous.position

    @property
    def distance(self) -> float:
        return self.position.distance_to(self.previous_position)

    @property
    def navigation_time(self):
        return self.previous_position.time_to(self.position, self.previous_departure_time)

    @property
    def waiting_time(self):
        return max(self.earliest - self.arrival_time, 0.0)

    @cached_property
    def arrival_time(self):
        return max(self.previous_departure_time + self.down_time + self.navigation_time, self.earliest)

    @cached_property
    def departure_time(self) -> float:
        return self.arrival_time + self.load_time

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('position', self.position),
            ('identifier', self.identifier),
        )

    def flush(self) -> None:
        for key in ('arrival_time', 'departure_time',):
            self.__dict__.pop(key, None)
        for planned_trip in self.planned_trips:
            planned_trip.flush()

    def flush_all_previous(self):
        self.flush()
        if self.previous is not None:
            self.previous.flush_all_previous()

    def merge(self, other: Stop) -> None:
        if self == other:
            return
        assert self.position == other.position

        self.extend_pickups(other.pickups)
        for planned_trip in other.pickups:
            planned_trip.pickup = self

        self.extend_deliveries(other.deliveries)
        for planned_trip in other.deliveries:
            planned_trip.delivery = self



from __future__ import annotations

import unittest
from copy import deepcopy
from typing import TYPE_CHECKING

import itertools as it
from uuid import UUID

import jinete as jit

from tests.utils import (
    generate_routes,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        List
    )


class TestPlanning(unittest.TestCase):
    routes: Set[jit.Route]
    loaded_routes: Set[jit.Route]
    planned_trips: List[jit.PlannedTrip]
    trips: List[jit.Trip]

    @classmethod
    def setUpClass(cls) -> None:
        cls.routes = generate_routes(3)
        cls.loaded_routes = set(route for route in cls.routes if route.loaded)
        cls.planned_trips = list(it.chain.from_iterable(route.planned_trips for route in cls.routes))
        cls.trips = list(planned_trip.trip for planned_trip in cls.planned_trips)

    def test_construction(self):
        planning = jit.Planning(self.routes)

        self.assertIsInstance(planning, jit.Planning)
        self.assertIsNotNone(planning.uuid)
        self.assertIsInstance(planning.uuid, UUID)
        self.assertIsNotNone(planning.routes)
        self.assertIsInstance(planning.routes, set)

        self.assertTrue(all(isinstance(route, jit.Route) for route in planning.routes))
        self.assertEqual(planning.routes, self.routes)

    def test_loaded_routes(self):
        planning = jit.Planning(self.routes)
        self.assertEqual(planning.loaded_routes, self.loaded_routes)

    def test_planned_trips(self):
        planning = jit.Planning(self.routes)
        self.assertEqual(list(planning.planned_trips), self.planned_trips)

    def test_trips(self):
        planning = jit.Planning(self.routes)
        self.assertEqual(list(planning.trips), self.trips)

    def test_as_tuple(self):
        planning = jit.Planning(self.routes)
        expected = (
            ('uuid', planning.uuid),
            ('route_identifiers', tuple(route.identifier for route in self.routes)),
        )
        self.assertEqual(expected, tuple(planning))

    def test_deepcopy(self):
        planning = jit.Planning(self.routes)

        copied_planning = deepcopy(planning)
        self.assertNotEqual(planning, copied_planning)
        self.assertNotEqual(id(planning), id(copied_planning))
        self.assertNotEqual(planning.uuid, copied_planning.uuid)
        self.assertEqual(len(planning.routes), len(copied_planning.routes))
        self.assertTrue(planning.routes.isdisjoint(copied_planning.routes))
        for route in planning.routes:
            assert any(set(route.trips).isdisjoint(copied_route.trips) for copied_route in copied_planning.routes)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
)

from .abc import (
    Model,
)

if TYPE_CHECKING:
    from typing import (
        Set,
        Generator,
        Tuple,
        Any,
        Type,
    )
    from uuid import (
        UUID,
    )
    from .trips import (
        Trip,
    )
    from .routes import (
        Route,
    )
    from .jobs import (
        Job,
    )
    from .vehicles import (
        Fleet,
        Vehicle,
    )
    from .plannings import (
        Planning,
    )
    from .objectives import (
        Objective,
        OptimizationDirection,
    )
    from ..algorithms import (
        Algorithm,
    )


class Result(Model):

    def __init__(self, fleet: Fleet, job: Job, algorithm_cls: Type[Algorithm], planning: Planning,
                 computation_time: float):
        self.fleet = fleet
        self.job = job
        self.algorithm_cls = algorithm_cls
        self.planning = planning
        self.computation_time = computation_time

    @property
    def trips(self) -> Set[Trip]:
        return self.job.trips

    @property
    def vehicles(self) -> Set[Vehicle]:
        return self.fleet.vehicles

    @property
    def routes(self) -> Set[Route]:
        return self.planning.routes

    @property
    def planning_uuid(self) -> UUID:
        return self.planning.uuid

    @property
    def feasible(self) -> bool:
        return self.planning.feasible

    @property
    def completed_trips(self) -> Set[Trip]:
        trips: Set[Trip] = set()
        for route in self.routes:
            trips |= set(route.loaded_trips)
        return trips

    @property
    def coverage_rate(self):
        return len(self.completed_trips) / len(self.job.trips)

    @property
    def objective(self) -> Objective:
        return self.job.objective

    @property
    def optimization_function(self) -> Tuple[float, ...]:
        return self.objective.optimization_function(self)

    @property
    def direction(self) -> OptimizationDirection:
        return self.objective.direction

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        yield from (
            ('fleet_uuid', tuple(self.fleet)),
            ('job', tuple(self.job)),
            ('algorithm_name', self.algorithm_cls.__name__),
            ('planning_uuid', self.planning_uuid)
        )



from pathlib import Path

import jinete as jit

BASE_PATH = Path(__file__).parents[1]
DATASETS_PATH = BASE_PATH / 'res' / 'datasets'


def main():
    file_path = DATASETS_PATH / 'hashcode' / 'a_example.in'
    loader = jit.FileLoader(file_path)

    print(f'Surface: \n  {loader.surface}')
    print(f'Fleet: \n  {loader.fleet}')
    print(f'Job: \n  {loader.job}')


if __name__ == '__main__':
    main()



from __future__ import annotations

import logging
from random import Random
from typing import (
    TYPE_CHECKING,
)

from ...models import (
    Route,
    Stop,
    PlannedTrip,
    Trip,
)

if TYPE_CHECKING:
    from typing import (
        Iterable,
        List,
        Union,
    )

logger = logging.getLogger(__name__)


class Conjecturer(object):

    def __init__(self, *args, **kwargs):
        pass

    def compute(self, route: Route, trips: Union[Trip, Iterable[Trip]], *args, **kwargs) -> List[Route]:
        if not isinstance(trips, Trip):
            return sum((self.compute(route, trip, *args, **kwargs) for trip in trips), [])
        trip = trips

        return [self.compute_one(route, trip, *args, **kwargs)]

    def compute_one(self, route: Route, trip: Trip, previous_idx: int = None, following_idx: int = None) -> Route:
        assert following_idx is None or (previous_idx is not None and following_idx is not None)

        if previous_idx is None:
            previous_idx = max(len(route.stops) - 2, 0)
        if following_idx is None:
            following_idx = max(len(route.stops) - 1, 0)

        route = route.clone(previous_idx + 1)

        pickup = self._build_pickup(route, trip, previous_idx)
        delivery = self._build_delivery(route, trip, previous_idx, following_idx, pickup)

        planned_trip = PlannedTrip(route.vehicle, trip, pickup, delivery)

        route.append_planned_trip(planned_trip)
        return route

    @staticmethod
    def _build_pickup(route: Route, trip: Trip, previous_idx: int) -> Stop:
        previous_pickup = route.stops[previous_idx]
        pickup = Stop(route.vehicle, trip.origin_position, previous_pickup)
        return pickup

    @staticmethod
    def _build_delivery(route: Route, trip: Trip, previous_idx: int, following_idx: int, pickup: Stop) -> Stop:
        previous_delivery = pickup if previous_idx + 1 == following_idx else route.stops[following_idx - 1]
        delivery = Stop(route.vehicle, trip.destination_position, previous_delivery)
        return delivery


class IntensiveConjecturer(Conjecturer):

    def compute(self, route: Route, trips: Union[Trip, Iterable[Trip]], *args, **kwargs) -> List[Route]:
        if not isinstance(trips, Trip):
            return super().compute(route, trips, *args, **kwargs)
        trip = trips

        routes = list()
        for i in range(len(route.stops) - 1):
            for j in range(i + 1, len(route.stops)):
                conjectured_route = self.compute_one(route, trip, i, j)
                routes.append(conjectured_route)
        return routes


class SamplingConjecturer(Conjecturer):
    def __init__(self, seed: int = 56, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.random = Random(seed)

    def compute(self, route: Route, trips: Union[Trip, Iterable[Trip]], count: int = 25,
                *args, **kwargs) -> List[Route]:
        if not isinstance(trips, Trip):
            return super().compute(route, trips, *args, **kwargs)
        trip = trips

        indices = set()
        for _ in range(count):
            sampled_i = self.random.randint(0, len(route.stops) - 2)
            sampled_j = self.random.randint(sampled_i + 1, len(route.stops) - 1)
            pair = (sampled_i, sampled_j)
            indices.add(pair)

        planned_trips = list()
        for i, j in indices:
            planned_trip = self.compute_one(route, trip, i, j)
            planned_trips.append(planned_trip)
        return planned_trips



import unittest

import jinete as jit


class TestMiscellaneous(unittest.TestCase):

    def test_version(self):
        self.assertIsInstance(jit.__version__, str)
        self.assertIsInstance(jit.VERSION, tuple)
        for value in jit.VERSION:
            self.assertIsInstance(value, int)
        self.assertEqual(jit.__version__, '.'.join(map(str, jit.VERSION)))


if __name__ == '__main__':
    unittest.main()



from .abc import (
    Crosser,
)
from .stateless import (
    StatelessCrosser,
)
from .best_stateless import (
    BestStatelessCrosser,
)
from .ordered import (
    OrderedCrosser,
)
from .randomized import (
    RandomizedCrosser,
)
